/*
 * XLife Copyright 1989 Jon Bennett jb7m+@andrew.cmu.edu, jcrb@cs.cmu.edu
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <pwd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "defs.h"

static float rnd_density = .9;

void announce(const char *s) {
/* display a string in the input window */
    if (convmode) {
         fprintf(stderr, "%s\n", s);
         exit(2);
    }
    else
    {
         (void) strcpy(inpbuf, s);
         DoExpose(INPUTWIN);
    }
}

void wait_activity(void) {
    if (!convmode)
         XMaskEvent(disp, KeyPressMask|ButtonPressMask, &event);
}

void fatal(const char *s) {
    fprintf(stderr, s);
    exit(-22);
}

int ClassifyWin(Window win) {
    if (win == lifew)
	return LIFEWIN;
    if (win == inputw)
	return INPUTWIN;
    if (win == mainw)
	return MAINWIN;
    if (win == rulew)
	return RULEWIN;
    if (win == coordw)
	return COORDWIN;
    if (win == helpw)
	return HELPWIN;
    return -1;
}

char *itoa(int n) {
    static char buf[16];
    char sign;
    char *ptr;

    if (n >= 0)
	sign = ' ';
    else {
	n = -n;
	sign = '-';
    }
    buf[15] = 0;
    ptr = buf + 14;
    do {
	*--ptr = n % 10 + '0';
	if (ptr <= buf)
	    return(buf);
	n /= 10;
    } while (n > 0);
    *--ptr = sign;
    return(ptr);
}

void drawcell(coord_t x, coord_t y, cell_t c) {
    if (scale <= 0)
	XDrawPoint(disp, lifew, cellgc[c], x, y);
    else {
	int sc;
	sc = (1 << scale) - (scale > 1);
	XFillRectangle(disp,lifew,cellgc[c],MOUSE2CELL(x),MOUSE2CELL(y),sc,sc);
    }
}

void drawpivot(coord_t x, coord_t y) {
    int sc;

    if (scale > 1) {
	sc = (1 << scale) - 1;
	XDrawLine(disp, lifew, xorgc, x, y, x + sc - 1, y + sc - 1);
	XDrawLine(disp, lifew, xorgc, x, y + sc - 1, x + sc - 1, y);
	XDrawPoint(disp, lifew, xorgc, x + sc/2, y + sc/2);
    }
    else
	XDrawPoint(disp, lifew, cellgc[1], x, y);
}

void drawbox(coord_t x1, coord_t y1, coord_t x2, coord_t y2, int color) {
/* box the area defined by two corners (life-world coordinates) */
    if (scale > 1) {
	int xbase, ybase, xedge, yedge, cellsize = 1 << scale, cc;

	xbase    = (((x1 > x2) ? x2 : x1) - xpos)*cellsize;
	ybase    = (((y1 > y2) ? y2 : y1) - ypos)*cellsize;
        xedge    = (((x1 > x2) ? x1 : x2) - xpos + 1)*cellsize;
        yedge    = (((y1 > y2) ? y1 : y2) - ypos + 1)*cellsize;
        /* X server works only with shorts */
        if (xbase < -32767) xbase = -32767; 
             else if (xbase > 32767) xbase = 32767;
        if (ybase < -32767) ybase = -32767;
             else if (ybase > 32767) ybase = 32767;
        if (xedge < -32767) xedge = -32767;
             else if (xedge > 32767) xedge = 32767;
        if (yedge < -32767) yedge = -32767;
             else if (yedge > 32767) yedge = 32767;
        /*
         * Actually, draw 2 boxes.  The second is two pixels wider.  This
         * compensates for the fact that a single-pixel vertical row is
         * near-invisible on many displays.
         */

        /* left-hand vertical line (doubled to cope with pixel aspect ratio) */
        if (color == 0 && dispmesh)
            cc = xbase%5 ? GRID_COLOR : GRID_ECOLOR;
        else
            cc = color;
        XDrawLine(disp, lifew, cellgc[cc], xbase - 1, ybase - 1,
               xbase - 1, yedge - 1);
        /*XDrawLine(disp, lifew, cellgc[cc], xbase - 2, ybase - 1,
               xbase - 2, yedge - 1);*/

        /* right-hand vertical line (doubled to cope with pixel aspect ratio)*/
        if (color == 0 && dispmesh)
            cc = xedge%5 ? GRID_COLOR : GRID_ECOLOR;
        else
            cc = color;
        XDrawLine(disp, lifew, cellgc[cc], xedge - 1, ybase - 1,
               xedge - 1, yedge - 1);
        /*XDrawLine(disp, lifew, cellgc[cc], xedge - 2, ybase - 1,
               xedge - 2, yedge - 1);*/

        /* top horizontal line */
        if (color == 0 && dispmesh)
            cc = ybase%5 ? GRID_COLOR : GRID_ECOLOR;
        else
            cc = color;
        XDrawLine(disp, lifew, cellgc[cc], xbase - 1, ybase - 1,
               xedge - 1, ybase - 1);

        /* bottom horizontal line */
        if (color == 0 && dispmesh)
            cc = yedge%5 ? GRID_COLOR : GRID_ECOLOR;
        else
            cc = color;
        XDrawLine(disp, lifew, cellgc[cc], xbase - 1, yedge - 1,
               xedge - 1, yedge - 1);
    }
}

void erasebox(coord_t x1, coord_t y1, coord_t x2, coord_t y2) {
/* actually, overwrite the box outline in black */
    if (scale > 1) {
	drawbox(x1, y1, x2, y2, 0);

	/*
	 * The process of erasing the old box may have messed up some
	 * cell edges in the box interior or around its outside edge. 
	 * Fix this by forcing a redisplay.
	 */
	redisplay();
    }
}

void showcoord(const bool force) {
/* update the coordinate & state display */
  static int f;
    /* Coordinate display added by Paul Callahan (callahan@cs.jhu.edu) */
    if (dispcoord) {
	coord_t x,y;
	char pstring[50];

	f = 1;
	x = XPOS(event.xmotion.x, xpos); 
	y = YPOS(event.xmotion.y, ypos); 

	if (force || lastx != x || lasty != y) {
	    lastx=x;
	    lasty=y; 
	    sprintf(pstring, "(%ld,%ld)=%d",
		    x - xorigin, y - yorigin,
		    lookcell(&active, x,y));
	    XClearWindow(disp,coordw);
	    XDrawString(disp,coordw,ntextgc,CWCOORDS(0,0),
			pstring,strlen(pstring));
	}
    }
    else if (f) {
        XClearWindow(disp,coordw);
        f = 0;
    }
}

void showrules(void) {
    char	buf[80], *cp;

    (void) strcpy(buf, active_rules);
    buf[0] = toupper(buf[0]);
    if (cp = strrchr(buf, '.')) *cp = '\0';
    XClearWindow(disp, rulew);
    XDrawString(disp, rulew, ntextgc, CWRULES(0,0), buf, strlen(buf));
}

void get_rnd_density() {
    float r;

    strcpy(inpbuf, "Input density (% or nothing): ");
    minbuflen = 31;
    DoExpose(INPUTWIN);
    getxstring();
    if (sscanf(inpbuf + 30, "%f", &r) == 1)
	rnd_density = r/100;
    sprintf(inpbuf, "Using %.2f%% density", rnd_density*100);
    DoExpose(INPUTWIN);
}

void randomize(void) {
    cellcount_t num = shr((unsigned)width, scale)*shr((unsigned)height, scale)*rnd_density;

    while (num--) {
	chgcell(&active,
		random()%shr((unsigned)width, scale) + xpos,
		random()%shr((unsigned)height, scale) + ypos, 1);
    }
    redisplay();
}

void settimeout(unsigned long ms) {
    sprintf(inpbuf, "Delay %d msec", delay);
    DoExpose(INPUTWIN);
    timeout.tv_sec = ms/1000;
    timeout.tv_usec = (ms%1000)*1000;
}

void stamp(char *leader, FILE *ofp) {
/* generate \n-terminated time and user info string */
    struct passwd *pw;
    char machine[80];
    time_t timeval;
    extern char *ctime();

    timeval = time(0);
    gethostname(machine, sizeof(machine));
    if ((pw = getpwuid(getuid())) != NULL)
        fprintf(ofp, "%s %s \"%s\"@%s %s",
                leader, pw->pw_name, pw->pw_gecos, machine,ctime(&timeval));
}

void benchmark(void) {
    char outbuf[100];
    u_long num, count;
    double tm;
    struct timeval start,end;
    struct timezone tmz;

    state = STOP;

    strcpy(inpbuf, "Number of generations: ");
    minbuflen = 23;

    DoExpose(INPUTWIN);
    getxstring();

    strcpy(outbuf, inpbuf + 23);
    sscanf(outbuf, "%ld", &count);

    gettimeofday(&start, &tmz);
    for (num = 0; num < count; num++)
       if (tentative.tiles)
           generate(&tentative);
       else
	   generate(&active);

    gettimeofday(&end, &tmz);
    tm = (end.tv_sec*1000000 + end.tv_usec - start.tv_sec*1000000 - start.tv_usec)/1000000.0;

    DoExpose(LIFEWIN);
    sprintf(inpbuf, "Time: %fs,  Press a key to continue", tm);
    DoExpose(INPUTWIN);
    wait_activity();

    displaystats();
}

#if STATEBITS > 1
void color_button(int location, int color, int large) {
    XWindowAttributes inattrs;

    XGetWindowAttributes(disp, inputw, &inattrs);
    XFillRectangle(disp, inputw, cellgc[color],
		   inattrs.width - BORDERWIDTH
			   - (maxstates - location)*(FONTWIDTH + 3)
			   - large,
		   BORDERWIDTH + 3 - large,
		   FONTWIDTH + large*2, FONTHEIGHT - 1 + large*2);
}

void setcolor(int val, unsigned long x, unsigned long y) {
    if (val == -1) {
	unsigned long left;
	XWindowAttributes inattrs;

	XGetWindowAttributes(disp, inputw, &inattrs);
	left = inattrs.width - BORDERWIDTH - maxstates * (FONTWIDTH + 3);

	if (x < left) return;

	val = (x - left) / (FONTWIDTH + 3);
    }

    color_button(paintcolor, 1, 2); 
    color_button(paintcolor, paintcolor, 0);
    color_button(val, 0, 2);
    color_button(val, val, 0);
    paintcolor = val;
}
#endif /* STATEBITS > 1 */
